#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fitsio.h"
#include "nrutil.h"

#define LINEPIX 1024
#define LINESCAN 1024
#define MAXDATA 1024
#define MAXCHAR 1024
#define MAXHDR 1024
#define HDRSIZE 80

main(int argc, char **argv)
{
  fitsfile *fp;
  FILE *fpscan;
  int i,j;
  int flag;
  int nscan;
  int istart,iend,imiddle;
  int masktype,offset;

  double a,b,c;
  double a1,b1,c1;
  double a2,b2,c2;
  double alpha;
  float **mat,**invmat;
  float *col,*vec;
  float d;
  int *indx;
  int nmatrix;

  int status,nfound,anynull,bitpix=-32;
  long fpixel, nelements,naxis=2,npixels;
  long naxes[2]={LINEPIX, LINEPIX},npix;
  float nullval; 
  char line[MAXCHAR];
  float *pix;
  float xx,yy;

  float *x, *y, *fsigma, *fpeak;
  float peak,sigma,peakmax,fwhm;
  float absdif;
  float ypos;

  char openfile[100],savefile[100];

  if(argc!=5 && argc!=6){
    fprintf(stderr,"Usage : %s scan_file output_file mask_type(gauss/square) offset(on/off/ohs) [fwhm] \n",argv[0]);
    exit(1);
  }

  sprintf(openfile, "%s",argv[1]);
  sprintf(savefile, "%s",argv[2]);
  if(strcmp(argv[3],"gauss")==0) masktype=0;
  else if(strcmp(argv[3],"square")==0) masktype=1;
  else {
    fprintf(stderr,"Mask type should be either `gauss` or `square`.\n");
    exit(1);
  }

  if(strcmp(argv[4],"on")==0) offset=1;
  else if(strcmp(argv[4],"off")==0) offset=0;
  else if(strcmp(argv[4],"ohs")==0) offset=-1;
  else {
    fprintf(stderr,"Offset parameter should be either `on` or `off`.\n");
    exit(1);
  }

  if(argc==6){
    fwhm=atof(argv[5]);
    sigma=fwhm/2.3548200 ;
  }
  
  
  /* allocate memory */
  x=(float *) malloc(LINESCAN*sizeof(float));
  y=(float *) malloc(LINESCAN*sizeof(float));
  fsigma=(float *) malloc(LINESCAN*sizeof(float));
  fpeak=(float *) malloc(LINESCAN*sizeof(float));

  /* allocate matrix */
  nmatrix=3;
  mat=matrix(1,nmatrix,1,nmatrix);
  invmat=matrix(1,nmatrix,1,nmatrix);
  indx=ivector(1,nmatrix);
  col=vector(1,nmatrix);
  vec=vector(1,nmatrix);

  /* load scan file */
  fprintf(stderr,"%s \t: Read %s  ... ",argv[0],openfile);
  if((fpscan=fopen(openfile,"r"))==NULL){
    fprintf(stderr,"Open Error  %s",openfile);
    exit(1);
  }
  i=0;
  nscan=0;
  if(fgets(line,MAXCHAR,fpscan)!=NULL){
    sscanf(line,"%d %d",&(naxes[0]),&(naxes[1]));
  }
  while(fgets(line,MAXCHAR,fpscan)!=NULL){
    sscanf(line,"%f %f %f %f",x,y,fsigma,fpeak);
    x++;
    y++;
    fsigma++;
    fpeak++;
    nscan++;
  }
  x-=nscan;
  y-=nscan;
  fsigma-=nscan;
  fpeak-=nscan;
  fclose(fpscan);
  fprintf(stderr,"OK\n");

  /* find pixel range to use */
  peakmax=fpeak[0];
  for(i=0;i<nscan;i++) if(peakmax<fpeak[i]) peakmax=fpeak[i];
  fprintf(stderr,"peak %f\n",peakmax);
  istart=-1;iend=-1;imiddle=-1;
  for(i=0;i<nscan;i++){
    if(istart==-1 && fpeak[i]>(peakmax/5)) istart=i;
    if(iend==-1 && istart!=-1 && x[i]>800 && fpeak[i]<(peakmax/5)) iend=i;
  }
  if(istart==-1) istart=0;
  if(iend==-1) iend=nscan;
  fprintf(stderr,"%f - %f\n",x[istart],x[iend-1]);
  for(i=istart;i<iend;i++){
    if(x[i]>512 && imiddle==-1) imiddle=i;
  }

  /* calculate average fwhm */
  if(argc==5){
    sigma=0.0;
    for(i=istart;i<iend;i++) sigma+=fsigma[i]/((float) (iend-istart));
    if(masktype==0) fwhm=sigma*2.3548200;
    else fwhm=sigma*2.3548200*3;
  }
  fprintf(stderr,"sigma %f\n",sigma);
  
  /*  fprintf(stderr,"%d %d %d\n",istart,imiddle,iend); */

  if(offset!=-1){
    /* fit */
    /* calculate matrix */
    for(j=1;j<=nmatrix;j++){
      vec[j]=0.0;
      for(i=1;i<=nmatrix;i++) mat[i][j]=0.0;
    }
    for(i=istart;i<imiddle;i++){
      /*    fprintf(stderr,"%f %f\n",x[i],y[i]); */
      if(fpeak[i]>(peakmax/10.0) && fpeak[i]>100.0){
	mat[1][1]+=pow(x[i],4);
	mat[1][2]+=pow(x[i],3);
	mat[2][1]+=pow(x[i],3);
	mat[1][3]+=pow(x[i],2);
	mat[2][2]+=pow(x[i],2);
	mat[3][1]+=pow(x[i],2);
	mat[3][2]+=x[i];
	mat[2][3]+=x[i];
	mat[3][3]+=1.0;
	
	vec[1]+=x[i]*x[i]*y[i];
	vec[2]+=x[i]*y[i];
	vec[3]+=y[i];
      }
    }
    
    /* calculate inverse */
    ludcmp(mat,nmatrix,indx,&d);
    for(j=1;j<=nmatrix;j++){
      for(i=1;i<=nmatrix;i++) col[i]=0.0;
      col[j]=1.0;
      lubksb(mat,nmatrix,indx,col);
      for(i=1;i<=nmatrix;i++) invmat[i][j]=col[i];
    }
    
    /* caluculate fit parameter */
    a1=0.0;b1=0.0;c1=0.0;
    for(i=1;i<=3;i++){
      a1+=(double) (invmat[i][1]*vec[i]);
      b1+=(double) (invmat[i][2]*vec[i]);
      c1+=(double) (invmat[i][3]*vec[i]);
    }
    
    /* calculate matrix */
    for(j=1;j<=nmatrix;j++){
      vec[j]=0.0;
      for(i=1;i<=nmatrix;i++) mat[i][j]=0.0;
    }
    for(i=imiddle;i<iend;i++){
      if(fpeak[i]>(peakmax/10.0) && fpeak[i]>100.0){
	mat[1][1]+=pow(x[i],4);
	mat[1][2]+=pow(x[i],3);
	mat[2][1]+=pow(x[i],3);
	mat[1][3]+=pow(x[i],2);
	mat[2][2]+=pow(x[i],2);
	mat[3][1]+=pow(x[i],2);
	mat[3][2]+=x[i];
	mat[2][3]+=x[i];
	mat[3][3]+=1.0;
	
	vec[1]+=x[i]*x[i]*y[i];
	vec[2]+=x[i]*y[i];
	vec[3]+=y[i];
      }
    }
    
    /* calculate inverse */
    ludcmp(mat,nmatrix,indx,&d);
    for(j=1;j<=nmatrix;j++){
      for(i=1;i<=nmatrix;i++) col[i]=0.0;
      col[j]=1.0;
      lubksb(mat,nmatrix,indx,col);
      for(i=1;i<=nmatrix;i++) invmat[i][j]=col[i];
    }
    
    /* caluculate fit parameter */
    a2=0.0;b2=0.0;c2=0.0;
    for(i=1;i<=3;i++){
      a2+=(double) (invmat[i][1]*vec[i]);
      b2+=(double) (invmat[i][2]*vec[i]);
      c2+=(double) (invmat[i][3]*vec[i]);
    }
    
    fprintf(stderr,"%e %e %e\n",a1,b1,c1);
    fprintf(stderr,"%e %e %e\n",a2,b2,c2);
    
    /* caluculate offset parameter */
    alpha=((a2-a1)*512.5*512.5+(b2-b1)*512.5+(c2-c1))/512.5;
    fprintf(stderr,"%e\n",alpha);
  }
    
  if(offset!=0){
    if(offset==1){
      /* add offset to plot*/
      for(i=0;i<imiddle;i++) y[i]+=alpha*x[i];
      for(i=imiddle;i<nscan;i++) y[i]+=alpha*(x[i]-512.5);
      /*    for(i=0;i<nscan;i++) fprintf(stdout,"%.0f %f %f %f\n",x[i],y[i],fsigma[i],fpeak[i]); */
    }

    /* final fit */
    /* calculate matrix */
    for(j=1;j<=nmatrix;j++){
      vec[j]=0.0;
      for(i=1;i<=nmatrix;i++) mat[i][j]=0.0;
    }
    for(i=istart;i<iend;i++){
      if(fpeak[i]>(peakmax/10.0) && fpeak[i]>100.0){
	mat[1][1]+=pow(x[i],4);
	mat[1][2]+=pow(x[i],3);
	mat[2][1]+=pow(x[i],3);
	mat[1][3]+=pow(x[i],2);
	mat[2][2]+=pow(x[i],2);
	mat[3][1]+=pow(x[i],2);
	mat[3][2]+=x[i];
	mat[2][3]+=x[i];
	mat[3][3]+=1.0;
	
	vec[1]+=x[i]*x[i]*y[i];
	vec[2]+=x[i]*y[i];
	vec[3]+=y[i];
      }
    }
    
    /* calculate inverse */
    ludcmp(mat,nmatrix,indx,&d);
    for(j=1;j<=nmatrix;j++){
      for(i=1;i<=nmatrix;i++) col[i]=0.0;
      col[j]=1.0;
      lubksb(mat,nmatrix,indx,col);
      for(i=1;i<=nmatrix;i++) invmat[i][j]=col[i];
    }
    
    /* caluculate fit parameter */
    a=0.0;b=0.0;c=0.0;
    for(i=1;i<=3;i++){
      a+=(double) (invmat[i][1]*vec[i]);
      b+=(double) (invmat[i][2]*vec[i]);
      c+=(double) (invmat[i][3]*vec[i]);
    }
    
    fprintf(stderr,"%e %e %e\n",a,b,c);
  }

  /* free matrix */
  free_matrix(mat,1,nmatrix,1,nmatrix);
  free_matrix(invmat,1,nmatrix,1,nmatrix);
  free_vector(col,1,nmatrix);
  free_vector(vec,1,nmatrix);
  free_ivector(indx,1,nmatrix);

  /* free memory */
  free(x);
  free(y);
  free(fsigma);
  free(fpeak);
  
  /* create mask frame */
  fprintf(stderr,"naxes1=%d 2=%d\n",naxes[0],naxes[1]);
  npixels=naxes[0]*naxes[1];
  pix=(float *) malloc(npixels*sizeof(float));

  fprintf(stderr,"fwhm %f\n",fwhm);
  for(i=0;i<naxes[1];i++){
    for(j=0;j<naxes[0];j++){
      xx=(float) j+1;
      yy=(float) i+1;
      if(offset==0){
	if(j<513){a=a1;b=b1;c=c1;}
	else{a=a2;b=b2;c=c2;}
      }
      if(masktype==0){
	pix[j+naxes[0]*i]=exp(-(yy-a*xx*xx-b*xx-c)*(yy-a*xx*xx-b*xx-c)/sigma/sigma/2.0);
	absdif=sqrt((yy-a*xx*xx-b*xx-c)*(yy-a*xx*xx-b*xx-c));
	if(absdif+0.5<=fwhm/2.0) ;
	else if(absdif-0.5<fwhm/2.0){
	  pix[j+naxes[0]*i]*=(fwhm/2.0-(absdif-0.5));
	}
	else pix[j+naxes[0]*i]*=0.0;
      }
      else{
	absdif=sqrt((yy-a*xx*xx-b*xx-c)*(yy-a*xx*xx-b*xx-c));
	if(absdif+0.5<=fwhm/2.0) pix[j+naxes[0]*i]=1;
	else if(absdif-0.5<fwhm/2.0){
	  pix[j+naxes[0]*i]=fwhm/2.0-(absdif-0.5);
	  /*	  fprintf(stderr,"%d\t%d\t%f\t%f\n",j,i,pix[j+naxes[0]*i],absdif); */
	}
	else pix[j+naxes[0]*i]=0;
      }
    }
  }
  ypos=a*512.5*512.5+b*512.5+c;
  fprintf(stderr,"a=%e b=%e c=%e\n",a,b,c);
  fprintf(stderr,"ypos=%f\n",ypos);
  
  /* create mask file */
  fprintf(stderr,"%s \t: Write %s  ... ",argv[0],savefile);
  unlink(savefile);

  status=0;
  if (fits_create_file(&fp, savefile, &status)){
    fits_report_error(stderr,status);
    return(status);
  }

  if (fits_create_img(fp, bitpix, naxis, naxes, &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  
  if (fits_write_key(fp, TSTRING, "MASKTYPE", argv[3], "Type of a spectrum mask" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TFLOAT, "MASKFWHM", &fwhm, "FWHM of the mask" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TFLOAT, "MASKSIGM", &sigma, "Sigma of the gaussian mask" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TFLOAT, "MASKYPOS", &ypos, "Y-position of mask at x=512.5" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TDOUBLE, "MASKA", &a, "Y=a*X*X+b*X+c" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TDOUBLE, "MASKB", &b, "Y=a*X*X+b*X+c" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  if (fits_write_key(fp, TDOUBLE, "MASKC", &c, "Y=a*X*X+b*X+c" , &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  
  fpixel=1;
  if (fits_write_img(fp, TFLOAT, fpixel, npixels, pix, &status)) 
    fits_report_error(stderr,status);
    
  if (fits_close_file(fp, &status)){
    fits_report_error(stderr,status);
    return(status);
  }
  fprintf(stderr,"OK\n");

  free(pix);
}
